use std::time::Instant;

use scraper::{Html, Selector};

const SAMPLES: [(&str, &str, &str); 3] = [
    ("tochb", "a[name]", include_str!("../html/tochb.html")),
    (
        "android",
        ".devsite-top-logo-row",
        include_str!("../html/android.html"),
    ),
    ("10mb", "b", include_str!("../html/10mb.html")),
];

fn main() {
    println!(concat!(
        "engine,page,",
        "parse min,parse avg,parse max,",
        "select min,select avg,select max,",
        "serial min,serial avg,serial max",
    ));
    run::<ScraperEngine>();
    run::<TlEngine>();
}

fn run<E: Engine>() {
    for (page_name, selector, content) in SAMPLES {
        print!("{},{}", E::NAME, page_name);

        // parse, build DOM
        let dom = {
            bench(20, || {
                E::parse(content);
            });
            E::parse(content)
        };

        // find elem
        bench(10000, || {
            dom.select(selector);
        });

        // serialize
        bench(10, || {
            assert!(!dom.serialize().is_empty());
        });

        println!();
    }
}

fn bench(repeats: u16, cb: impl Fn()) {
    let mut min = f64::INFINITY;
    let mut max = 0.0;
    let mut sum = 0.0;
    for _ in 0..repeats {
        let start = Instant::now();
        cb();
        let delta = Instant::now() - start;

        let millis = delta.as_nanos() as f64 / 1_000_000.0;

        if millis < min {
            min = millis;
        }
        if millis > max {
            max = millis;
        }

        sum += millis;
    }
    print!(",{:.5},{:.5},{:.5}", min, sum / repeats as f64, max);
}

trait Engine {
    const NAME: &'static str;
    fn parse(doc: &str) -> impl Document;
}

trait Document<'d> {
    fn select(&self, pattern: &str);
    fn serialize(&self) -> String;
}

pub struct ScraperEngine;
impl Engine for ScraperEngine {
    const NAME: &'static str = "html5ever";

    fn parse(doc: &str) -> impl Document<'_> {
        ScraperDocument(Html::parse_document(doc))
    }
}

pub struct ScraperDocument(Html);
impl Document<'_> for ScraperDocument {
    fn select(&self, pattern: &str) {
        self.0
            .select(&Selector::parse(pattern).unwrap())
            .next()
            .unwrap();
    }

    fn serialize(&self) -> String {
        self.0.html()
    }
}

pub struct TlEngine;
impl Engine for TlEngine {
    const NAME: &'static str = "tl";

    fn parse(doc: &str) -> impl Document<'_> {
        TlDocument(tl::parse(doc, tl::ParserOptions::default()).unwrap())
    }
}

pub struct TlDocument<'d>(tl::VDom<'d>);
impl Document<'_> for TlDocument<'_> {
    fn select(&self, pattern: &str) {
        self.0.query_selector(pattern).unwrap().next().unwrap();
    }

    fn serialize(&self) -> String {
        self.0.outer_html()
    }
}
